#Copyright Daniel Dunn 2020
#This file is part of Scullery.

#Scullery is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, version 3.

#Scullery is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with Scullery.  If not, see <http://www.gnu.org/licenses/>.


_ureg = None
import math
#Units datafile.  All the units are all mixed together.  Every type of unit has to have a single base unit
#and everything else is defined in terms of that
units ={

    #Base unit of mass is grams
    "g": 1.0,
    "lb": 453.592,
    "oz": 28, 

    #Base unit of distance is meters
    "m": 1.0,
    "in": 0.0254,

    #Base unit of temperature is Kelvin
    "K": 1.0,
    #Offset unit. It is defined by two functions,the one to convert to the base an one to convert from
    "degC": (lambda x: x+273.15, lambda x: x-273.15),

    #Voltage/Current
    "A": 1,
    "V": 1,

    #Power, base is W
    "W": 1,
    "dBm": (lambda x: (10**(x/10))/1000, lambda x: 10*math.log10(x/0.001)),
    
    #Nobody really uses the uno I don't think but it's the only namw we have
    "U":1,
    "uno": 1,
    "%": 0.01,
    "ppm": 100/10**6,
    "ppb": 100/10**9,
    "dB":  (lambda x: 10**(x*10), lambda x: 10*math.log10(x)),
    
    'Pa': 1,
    'psi': 6894.76,
    'PSI': 6894.76
}

unitTypes={
     #Base unit of mass is grams
    "g":  "mass",
    "lb": "mass",
    "oz": "mass",

    #Base unit of distance is meters
    "m":  "length",
    "in": "length",

    #Base unit of temperature is Kelvin
    "K":    "temperature",
    #Offset unit. To convert FROM the base divide by the first and add the second number
    "degC": "temperature",

    "V": "voltage",
    "A": "current",
    "W": "power",
    "dBm": "power",

    "%": "ratio",
    "ppm":"ratio",
    "ppb":"ratio",
    "U": "ratio",
    "uno": "ratio",

    'psi': 'pressure',
    'Pa': 'pressure',
    "PSI": 'pressure'
}

_prefixes = {"n":10**-9,"u":10**-6, "m":0.001, "k":1000,"M":1000000,"G":1000000000,"T":1000000000000}
def getUnitType(u):
    return unitTypes(u)

def parsePrefix(u):
    if u in units:
        return(units[u],1)
    elif u[0] in _prefixes and u[1:] in units:
        if not isinstance(units[u[1:]], (int, float)):
            raise ValueError("SI prefix not supported with nonlinear units")
        return (units[u[1:]], _prefixes[u[0]])
    
    raise KeyError("No such unit: "+u)

def defineUnit(unitname, multiplier, type, offset=0,base=None):
    """Define a new unit. The multiplier is how many of the base unit one step in the base unit new unit is worth. 
    Base defaults to 1, the global base unit.
    
    Multiplier can also be a tuple of two functions, frombase,tobase, that handle the conversions. 
    Function based units cannot be chained. You must define them in terms of the global default base
    
    """

    if base:
        z = units[base]
        if isinstance(z,(int,float)):
            multiplier*= z
        else:
            raise RuntimeError("Function-based units can only be defined directly in terms of the global base")
        
        if not isinstance(multiplier,(int,float)):
            raise RuntimeError("Function-based units can only be defined directly in terms of the global base")

    if isinstance(multiplier,(int,float)):
        multiplier = multiplier
    units[unitname]= multiplier
    unitTypes[unitname]= type

defineUnit("degF", (lambda x: (x+459.67) * (5/9),  lambda x: (x*(9/5))-459.67), "temperature")

defineUnit("ft", 12, "length", 0, "in")
defineUnit("mile",1609.344, "length")

defineUnit("m3/min",1, "flow")
defineUnit("cfm", 0.028316847, "flow")
defineUnit("gpm", 0.0037854118, "flow")

defineUnit("Pa", 1,"pressure")
defineUnit("psi", 6894.7573, "pressure")
defineUnit("ksi", 1000,"pressure",0,"psi")
defineUnit("mmHg",133.32237,"pressure")

def _loadPint():
    global _ureg
    if not _ureg:
        import pint 
        _ureg = pint.UnitRegistry()

def convert(v, fromUnit, toUnit):
    if fromUnit==toUnit:
        return v
    try:
        fr,frmul = parsePrefix(fromUnit)
        to,tomul = parsePrefix(toUnit)
    #Fallback to pint library
    except KeyError:
        if not _ureg:
            _loadPint()
        x = _ureg.Quantity(v, fromUnit)
        return x.to(toUnit).magnitude

    v*=frmul
    #Convert into the base unit
    if isinstance(fr, (float,int)):
        v=v*fr
    else:
       v = fr[0](v)


    #Convert from the base unit
    if isinstance(to, (float,int)):
        v=v/to
    else:
        v=to[1](v)
        
    return v/tomul


def siFormatNumber(number,digits=2):
    if number == 0:
        return "0"
    if number > 10**15:
        return(str(iround(number/1000000000000000.0,digits))+'P')
    if number > 10**12:
        return(str(iround(number/1000000000000.0,digits))+'T')
    if number > 1000000000:
        return(str(iround(number/1000000000.0,digits))+'G')
    if number > 1000000:
        return(str(iround(number/1000000.0,digits))+'M')
    if number > 1000:
        return(str(iround(number/1000.0,digits))+'K')
    if number < 10**-12:
        return(str(round(number*(10**-15),digits))+'f')
    if number < 10**-9:
        return(str(round(number*1000000000000.0,digits))+'p')
    if number <10**-6:
        return(str(iround(number*1000000000.0,digits))+'n')
    if number < 0.001:
        return(str(iround(number*1000000.0,digits))+'u')
    if number < 0.5:
        return(str(iround(number*1000.0,digits))+'m')
    return str(iround(number,digits))

time_as_seconds ={
'year' : 60*60*24*365,
#A "month" as commonly used is a vauge unit. is it 28 days? 30? 31?
#To solve that, I define it as 1/12th of a solar year.
'month' : 60*60*24*30.4368333333,
'week' : 60*60*24*7,
'day' : 60*60*24,
"hour" : 60*60,
"minute" : 60,
"second" : 1,
"millisecond" : 0.001,
"microsecond" : 0.000001,
"nanosecond"  : 0.000000001,
"picosecond"  : 0.000000000001,
"femtosecond"  :0.000000000000001,
}


time_as_seconds_abbr ={
'yr' : 60*60*24*365,
#A "month" as commonly used is a vauge unit. is it 28 days? 30? 31?
#To solve that, I define it as 1/12th of a solar year.
'month' : 60*60*24*30.4368333333,
'w' : 60*60*24*7,
'd' : 60*60*24,
"h" : 60*60,
"m" : 60,
"s" : 1,
"ms" : 0.001,
"us" : 0.000001,
"ns"  : 0.000000001,
"ps"  : 0.000000000001,
"fs"  :0.000000000000001,
}
def timeIntervalFromString(s):
    """Take a string like '10 hours' or 'five minutes 32 milliseconds'
    or '1 year and 1 day' to a number of seconds"""
    regex = r"([0-9]*)\D*?(year|month|week|day|hour|minute|second|millisecond)s?"
    r = re.compile(regex)
    m = r.finditer(s)
    total = 0
    for i in m:
        multiplier= time_as_seconds[i.group(2).strip()]
        number = float(i.group(1))
        total += number*multiplier
    return total


def formatTimeIntervallong(t,maunitss,clock=False):
    """Take a length of time t in seconds, and return a nicely formatted string
    like "2 hours, 4 minutes, 12 seconds".
    maunitss is the maximum number of units to use in the string(7 will add a milliseconds field to times in years)

    """
    if clock:
        frac = t%1
        t -= frac
        seconds = t%60
        t-= seconds
        minutes = (t-(int(t/3600)*3600))/60
        t -= t%3600
        hours= t/3600

        s = "%02d:%02d"%(hours,minutes)
        if maunitss>2:
            s+= ":%02d"%(seconds)
        if maunitss>3:
            #Adding 0.01 seems to help with some kind of obnoxious rounding bug thing. Prob a better way to do things.
            s+=":%03d"%(0.01+frac*1000)

        return s
    s = ""
    
    for i in sorted(time_as_seconds.items(),key= lambda x:x[1], reverse=True):
        if maunitss == 0:
            return s[:-2]
        x = t%i[1]
        b=(t-x)
        y=(t-x)/i[1]

        t = t-b
        if y>1:
            s += str(int(round(y))) + " " + i[0]+"s, "
            maunitss -=1
        elif y==1:
            s += str(int(round(y))) + " " + i[0]+", "
            maunitss -=1
    return s[:-2]

def formatTimeIntervalabbr(t,maunitss,clock=False):
    """Take a length of time t in seconds, and return a nicely formatted string
    like "2 hours, 4 minutes, 12 seconds".
    maunitss is the maximum number of units to use in the string(7 will add a milliseconds field to times in years)

    """
    if clock:
        frac = t%1
        t -= frac
        seconds = t%60
        t-= seconds
        minutes = (t-(int(t/3600)*3600))/60
        t -= t%3600
        hours= t/3600

        s = "%02d:%02d"%(hours,minutes)
        if maunitss>2:
            s+= ":%02d"%(seconds)
        if maunitss>3:
            #Adding 0.01 seems to help with some kind of obnoxious rounding bug thing. Prob a better way to do things.
            s+=":%03d"%(0.01+frac*1000)

        return s
    s = ""
    for i in sorted(time_as_seconds_abbr.items(),key= lambda x:x[1], reverse=True):
        if maunitss == 0:
            return s[:-1]
        x = t%i[1]
        b=(t-x)
        y=(t-x)/i[1]

        t = t-b
        if y>1: 
            if i[0]=="month":
                s += str(int(round(y))) + i[0]+"s "
            else:
                s += str(int(round(y))) + i[0]+" "
            maunitss -=1
        elif y:
            s += str(int(round(y))) + i[0]+" "
            maunitss -=1
    return s[:-1]

formatTimeInterval = formatTimeIntervalabbr

def strToIntWithSIMultipliers(s):
    """Take a string of the form number[k|m|g] or just number and convert to an actual number
    '0'-> 0, '5k'->5000 etc. Does not do division!!!! m is mega not milli!!!"""
    s = s.lower()
    #This piece of code interprets a string like 89m or 50k as a number like 89 millon or 50,000
    if s.endswith('k'):
        return int(s[:-1])*1000
    elif s.endswith('m'):
        return int(s[:-1])*1000000
    elif s.endswith('g'):
        return int(s[:-1])*1000000000
    else:
        return int(s[:-1])

def iround(number,digits):
    if (number-int(number)==0) or (digits==0):
        return int(number)
    else:
        return round(number,digits)



#Copyright Daniel Dunn 2020 

#Scullery is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, version 3.

#Scullery is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with Scullery.  If not, see <http://www.gnu.org/licenses/>.

import time,weakref,os,yaml,threading
gmInstruments = None

players = weakref.WeakValueDictionary()

lock = threading.Lock()

def allNotesOff():
    try:
        for i in players:
            players[i].fs.all_notes_off(-1)
    except:
        pass

def stopAll():
    try:
        for i in players:
            players[i].close()
    except:
        pass

def remakeAll():
    try:
        for i in players:
            players[i].close()
    except:
        pass

def getGMInstruments():
    global gmInstruments
    if gmInstruments:
        return gmInstruments
    with open(os.path.join(os.path.dirname(os.path.abspath(__file__)),'gm_instruments.yaml')) as f:
        gmInstruments = yaml.load(f.read())
    return gmInstruments

def findGMInstrument(name, look_in_soundfont=None,bank=None):
    #Allow manually selected instruments
    try:
        return (bank,int(name))
    except:
        pass
    name = name.replace("(",'').replace(')','')
    #Try to find a matching patch name in the soundfont
    if look_in_soundfont:
        try:
            from sf2utils.sf2parse import Sf2File
            with open(look_in_soundfont, 'rb') as sf2_file:
                sf2 = Sf2File(sf2_file)
            #Names
            x= [i[0].split(b"\0")[0].decode("utf8") for i in sf2.raw.pdta['Phdr']]
        except Exception as e:
            logging.exception("Can't get metadata from this soundfont")
            print("Error looking through soundfont data",e)
            x = []

        #Indexes
        for i in range(len(x)):
            n = x[i].lower()
            n=n.replace("(",'').replace(')','')
            n=n.split(" ")            
            match = True

            for j in name.lower().split(" "):
                if not j in n:
                    #Bank 128 is reserved for drums so it can substitute for drum related words,
                    #It's still a match
                    if not (j in ('kit','drums','drum') and sf2.raw.pdta['Phdr'][i][2]==128):
                        match = False
            if match:
                if bank==None:
                    return (sf2.raw.pdta['Phdr'][i][2],sf2.raw.pdta['Phdr'][i][1])
                return (bank,sf2.raw.pdta['Phdr'][i][1])

    x= getGMInstruments()
    for i in x:
        n = x[i].lower()
        n=n.replace("(",'').replace(')','')
        n=n.split(" ")
        match = True
        for j in name.lower().split(" "):
            if not j in n:
                match = False
        if match:
            return (bank or 0,i)
    raise ValueError("No matching instrument")

def waitForJack():
    from . import jack
    for i in range(10):
        if not jack.getPorts():
            time.sleep(1)
        else:
            return

    raise RuntimeError("It appears that JACK is not running")

class FluidSynth():
    defaultSoundfont = "/usr/share/sounds/sf2/FluidR3_GM.sf2"

    def __init__(self, soundfont=None,jackClientName=None,
        connectMidi=None, connectOutput=None,reverb=False,chorus=False, ondemand=True):
        players[id(self)]=self

        if jackClientName:
           from . import jack
           waitForJack()
        
        self.soundfont = soundfont or self.defaultSoundfont

        if not os.path.isfile(self.soundfont):
            raise OSError("Soundfont: "+soundfont+" does not exist or is not a file")

        from . thirdparty import fluidsynth
        def remake():
            self.fs = fluidsynth.Synth()
            self.fs.setting("synth.chorus.active", 1 if chorus else 0)
            self.fs.setting("synth.reverb.active", 1 if reverb else 0)

            try:
                self.fs.setting("synth.dynamic-sample-loading", 1 if ondemand else 0)
            except:
                logging.exception("No dynamic loading support, ignoring")
            self.sfid = self.fs.sfload(self.soundfont)
            usingJack = False

            if jackClientName:
                self.fs.setting("audio.jack.id", jackClientName)
                self.fs.setting("audio.midi.id", "KaithemFluidsynth")

            usingJack = True

            if connectMidi:
                pass
                #self.midiAirwire = jackmanager.Mono

            if connectOutput:
                self.airwire= jack.Airwire(jackClientName or 'KaithemFluidsynth',connectOutput)
                self.airwire.connect()

            if usingJack:
                if not jackClientName:
                        self.fs.setting("audio.jack.id", "KaithemFluidsynth")
                        self.fs.setting("audio.midi.id", "KaithemFluidsynth")


                self.fs.setting("midi.driver", 'jack')
                self.fs.start(driver="jack", midi_driver="jack")

            else:
                #self.fs.setting("audio.driver", 'alsa')
                self.fs.start()
            for i in range(16):
                self.fs.program_select(i, self.sfid, 0, 0)
        remake()

        #allow restart after JACK settings change
        self.remake = remake

    def setInstrument(self,channel, instrument,bank=None):
        bank, insNumber = findGMInstrument(instrument,self.soundfont,bank)
        self.fs.program_select(channel, self.sfid, bank, insNumber)

    def noteOn(self,channel,note, velocity):
        self.fs.noteon(channel, note, velocity)

    def noteOff(self,channel,note):
        self.fs.noteoff(channel, note)
    def cc(self,channel,control, val):
        self.fs.cc(channel, control, val)

    def pitchBend(self,channel,val):
        self.fs.pitch_bend(channel, val)

    def programChange(self,channel,val):
        self.fs.program_change(channel, val)

    def __del__(self):
        self.close()

    def close(self):
        with lock:
            try:
                if hasattr(self,'fs'):
                    self.fs.delete()
                    del self.fs
            except AttributeError:
                pass





import unittest
from scullery import mnemonics
from scullery import units

class TestMisc(unittest.TestCase):
    def test_wordlist(self):
        self.assertEqual(len(mnemonics.wordlist),1633)
            
    def test_hash(self):
        "This just happens to be what you get if you hash that string"
        self.assertEqual(mnemonics.memorableBlakeHash(b'',3,''), 'initialpuzzlegroup')

            
    def test_hash8(self):
        "Test with a longer requested result"
        self.assertEqual(mnemonics.memorableBlakeHash(b'',8,''), 'initialpuzzlegroupdeclarefactoragendalorenzoacademy')
  

    def test_conversions(self):
        "Test with a longer requested result"
        self.assertEqual(units.convert(12,"in","ft"), 1)
        self.assertEqual(units.convert(1,"km","m"), 1000)
        self.assertAlmostEqual(units.convert(50,"degC","degF"), 122)
        self.assertAlmostEqual(units.convert(122,"degF","degC"), 50)
        self.assertAlmostEqual(units.convert(1,"kft","in"), 12000)
        #Milli-inches?
        self.assertAlmostEqual(units.convert(1,"ft","min"), 12000)


    def test_nonsense_convert(self):
        with self.assertRaises(Exception):
            units.convert("80","trashpiles","garbages")
        #Can't use SI prefixes with nonlinear units like degrees
        with self.assertRaises(Exception):
            units.convert("80","mdegC","K")


  


from . import iceflow, jack

def volumeToDB(vol):
    if vol<0.001:
        return -100
    #Calculated usiung curve fitting, assuming that 0 is 0db,
    #0.5 is 10db below, etc.
    return -38.33333 + 77.80645*vol- 39.56989*vol**2

def volumeToRawGain(vol):
    vGain = 10**(volumeToDB(vol)/20)

class AudioFilePlayer(iceflow.GStreamerPipeline):
    def __init__(self, file, volume=1, output="@auto",onBeat=None, _prevPlayerObject=None):

        if output==None:
            output="@auto"
        self.filename = file


        self.lastFileSize = os.stat(file).st_size

        if self.lastFileSize==0:
            for i in range(10):
                self.lastFileSize = os.stat(file).st_size
                if self.lastFileSize>0:
                    break
                time.sleep(1)

        if self.lastFileSize==0:
            raise RuntimeError("File was still zero bytes after 10 seconds")
        
        self.lastPosition = 0
       
        self.lastFileSizeChange = 0

        gstwrapper.Pipeline.__init__(self,str(uuid.uuid4()),systemTime=True,realtime=70)
        self.aw =None
        self.ended = False

        self.lastBeat = 0
        self.peakDetect = 0
        self.detectedBeatInterval = 1/60
        self.beat = onBeat
        
        self._prevPlayerObject = _prevPlayerObject


        self.src = self.addElement('filesrc',location=file)
        self.queue = self.addElement("queue")

        decodeBin = self.addElement('decodebin',low_percent=15)
        #self.addElement('audiotestsrc')
        isVideo=False

        for i in (".mkv",".flv",".wmv",".mp4",".avi"):
            if file.endswith(i):
                isVideo=True
        if isVideo:
            decodeBin.set_property("low-percent",80)
            q=self.addElement('queue', connectToOutput=decodeBin, connectWhenAvailable="audio")
        
            self.addElement('audiorate')
            self.addElement('audioconvert')

        else:
            self.addElement('audioconvert',connectToOutput=decodeBin,connectWhenAvailable="audio")
    
        self.addElement('audioresample')


        
        if onBeat:
            self.addLevelDetector()

        self.fader = self.addElement('volume', volume=volume)

        if output=="@auto":
            self.sink = self.addElement('autoaudiosink')

        elif output.startswith("@alsa:"):
            self.addElement('alsasink',device= output[6:])

        elif output.startswith("@pulse"):
            self.addElement('alsasink',device= output[6:])

        #No jack clients at all means it probably isn't running
        elif not jack.getPorts():
            self.addElement('alsasink',device= output)
        
        #Default to just using jack
        else:
            cname="player"+str(time.monotonic())+"_out"

            self.sink = self.addElement('jackaudiosink', buffer_time=8000 if not isVideo else 80000, 
            latency_time=4000 if not isVideo else 40000,slave_method=2,port_pattern="jhjkhhhfdrhtecytey",
            connect=0,client_name=cname)

            self.aw = jackmanager.Airwire(cname, output)
            self.aw.connect()
        #Get ready!
        self.pause()


    def loopCallback(self):
        size= os.stat(self.filename).st_size
        if not size==self.lastFileSize:
            self.lastFileSizeChange = time.monotonic()
            self.lastFileSize=size

        self.lastPosition = self.getPosition()

    def onEOS(self):
        #If the file has changed size recently, this might not be a real EOS,
        #Just a buffer underrun. Lets just try again 
        if self.lastFileSizeChange> (time.monotonic()-3):
            self.pause()
            self.seek(self.lastPosition-0.1)
            time.sleep(3)
            self.play()
        else:
            gstwrapper.Pipeline.onEOS(self)


    def setFader(self,level):
        "Low level function that sets the raw volume element"
        with self.lock:
            try:
                if self.fader:
                    self.fader.set_property('volume', level)
            except ReferenceError:
                pass


    def setVolume(self,v):
            #We are going to do a perceptual gain algorithm here
            self._volume=v
            #Now convert to a raw gain
            vGain = volumeToRawGain(v)
            self.setFader(vGain if not vGain<-0.01 else 0)

    @property
    def volume(self):
        return self._volume

    @volume.setter
    def volume(self, volume):
        self.setFader(volume)
    
    def stop(self):

        #Prevent any ugly noise that GSTreamer might make when stopped at
        #The wrong time
        try:
            if self.aw:
                self.aw.disconnect()
        except:
            logging.exception("Err disconnecting airwire")

        gstwrapper.Pipeline.stop(self)

        if self._prevPlayerObject:
            p = self._prevPlayerObject()
            if p:
                p.stop()

    def resume(self):
        self.start()
  
    def onStreamFinished(self):
        self.ended=True

    def isPlaying(self):
        return self.running


    def addLevelDetector(self):
        self.addElement("level", post_messages=True, peak_ttl=300*1000*1000,peak_falloff=60)

    def on_message(self, bus, message,userdata):
        s = message.get_structure()
        if not s:
            return True
        if  s.get_name() == 'level':
            if self.board:
                rms = sum([i for i in s['rms']])/len(s['rms'])
                self.peakDetect = max(self.peakDetect, rms)
                timeSinceBeat = time.monotonic()-self.lastBeat

                threshold =self.peakDetect-(1+ (3*max(1,timeSinceBeat/self.detectedBeatInterval)))
                if timeSinceBeat> self.detectedBeatInterval/8:
                    if rms>threshold:
                        self.beat()
                        self.detectedBeatInterval = (self.detectedBeatInterval*3 + timeSinceBeat)/4
                        self.peakDetect*=0.996
        return True





import unittest,gc,time,os,random

from scullery import workers
workers.start()

from scullery import messagebus

def workerTest(extra=0):
    l = []
    def f():
        l.append(1)
    workers.do(f)
    for i in range(200000):
        if l:
            break
        time.sleep(0.00001)
    time.sleep(extra)
    return len(l)



def workerSpam(count=10):
    l = []
    def f():
        l.append(1)
    for i in range(count):
        workers.do(f)

assert(workerTest())

class TestMsgbus(unittest.TestCase):
    def test_sync(self):
        p = []
        def f(t,v):
            p.append((t,v))
        messagebus.subscribe("/test",f)
        messagebus.postMessage("/test","foo",synchronous=True)

        self.assertEqual(p,[('/test','foo')])

    def test_del(self):
        "Test unsubscribing by deleting a function"
        p = []
        def f(v):
            p.append(v)
        messagebus.subscribe("test",f)
        del f
        gc.collect()

        messagebus.postMessage("test","foo",synchronous=True)

        self.assertEqual(p,[])   

    def test_async(self):
        p = []
        def f(v):
            p.append(v)
        messagebus.subscribe("test",f)
        messagebus.postMessage("test","foo")

        s = 100
        while (not p) and s:
            time.sleep(0.01)
            s-=1
        self.assertEqual(p,['foo'])

    def test_Workers(self):
        workerSpam(100)
        self.assertLessEqual(len(workers.workers),workers.maxWorkers)

        for i in range(1000):
            self.assertEqual(workerTest(),1)

        for i in range(100):
            self.assertEqual(workerTest(),1)
            time.sleep(0.001)

        for i in range(1):
            self.assertEqual(workerTest(0.1),2)

        for i in range(100):
            self.assertEqual(workerTest(),1)
            time.sleep(random.random()*0.001)
        
        self.assertLessEqual(len(workers.workers),workers.maxWorkers)

        time.sleep(3)
        #Workers should stop within 1s of inactivity.
        self.assertEqual(len(workers.workers),workers.minWorkers)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="scullery",
    version="0.1.5",
    author="Daniel Dunn",
    author_email="dannydunn@eternityforest.com",
    description="A utility library based on KaithemAutomation featuring a GStreamer wrapper",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/EternityForest/scullery",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent"
    ],
    python_requires='>=3.6',
    install_requires=[
          'pyyaml',
          'typeguard',
          "sf2utils",
          "pyFluidSynth",
          "paho-mqtt"
      ],
)



import hashlib, struct,os

wordlist = [s.strip() for s in open(os.path.join(os.path.dirname(os.path.abspath(__file__)),'words_mnemonic.txt'))]

def memorableBlakeHash(x, num=3, separator=""):
    "Use the diceware list to encode a blake hash of x. This IS meant to be secure-ish(For long enough hashes)"
    o = ""

    if isinstance(x, str):
        x = x.encode("utf8")
    
    hash = hashlib.blake2b(x).digest()

    x=hash

    for i in range(num):
        # 1024 is an even divisor of 2**16
        n = struct.unpack("<H",x[:2])[0]
        
        #We are going to rehash here, to avoid any modulo bias
        while n>2**16 - len(wordlist):
            hash = hashlib.blake2b(hash).digest()
            x=hash
            n = struct.unpack("<H",x[:2])[0]

        n = n%len(wordlist)

        o+= wordlist[n] + separator
        x=x[2:]   
    return o[:-len(separator)] if separator else o



#Copyright Daniel Dunn 2019
#This file is part of Scullery.

#Scullery is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, version 3.

#Scullery is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with Scullery.  If not, see <http://www.gnu.org/licenses/>.
__doc__=''



import weakref
import threading
import base64

import os,re,time,subprocess,hashlib,struct,threading,atexit,select,traceback

#Util is not used anywhere else
from . import workers,mnemonics,util
wordlist = mnemonics.wordlist

#This is an acceptable dependamcy, it will be part of libkaithem if such a thing exists
from . import messagebus, iceflow

def shouldAllowGstJack(*a):
    global _jackclient
    for i in range(0,10):
        #Repeatedly retry, because this might just be a case
        #Of jack not being ready at boot yet.
        with lock:
            try:
                t=getPorts()
                if  t:
                    return True
            except:
                if i>8:
                    print(traceback.format_exc())
            time.sleep(1)

    return False


iceflow.GstreamerPipeline.shouldAllowGstJack = shouldAllowGstJack

import logging

log = logging.getLogger("system.jack")

_jackclient =None


lock = threading.RLock()


def onJackFailure():
    pass

def onJackStart():
    pass


prevJackStatus=False


#Default settings
jackPeriods = 3
periodSize = 512
jackDevice = "hw:0,0"

useAdditionalSoundcards = 'yes'

#Make sure the alsa_in manager knows what we select as the primary, so that
#It doesn't get stepped on
currentPrimaryDevice = "hw:0,0"
#These apply to soundcards other than the main system card
usbPeriodSize = -1
usbPeriods = -1

usbLatency = -1
usbQuality = 0

realtimePriority = 70

#Do we want to run PulseAudio and the pulse jack backend?
usePulse= True

sharePulse = None

#Should we create alsa_in and alsa_out ports for every soundcard, with persistant names?
manageSoundcards = True

#Should we auto restart the jack process?
#No by default, gets auto set to True by startJackProcess()
manageJackProcess = False




def settingsReloader():
    pass


#Currently we only support using the default system card as the
#JACK backend. We prefer this because it's easy to get Pulse working right.
usingDefaultCard = True

def isConnected(f,t):
    if not isinstance(f, str):
        f=f.name
    if not isinstance(t, str):
        t=t.name

    if (t,f) in _realConnections:
        return True
    if (f,t) in _realConnections:
        return True

def setupPulse():

    log.debug("Setting pulseaudio integration for JACK")
    try:
        subprocess.call(['pulseaudio','-k'],timeout=5,stderr=subprocess.DEVNULL,stdout=subprocess.DEVNULL)
    except:
        print(traceback.format_exc())

    if not usePulse:
        return


    if not util.which("pulseaudio"):
        logging.error("Pulseaudio requested but not installed")
        return


    time.sleep(0.1)
    #-n prevents loading the defaults, which would otherwise
    #Cause trouble with pulseaudio
    cmd = ['pulseaudio',"-D", "-n","--file="+os.path.join(os.path.dirname(__file__),"pulse_jack_rc.txt")]

    if sharePulse=="system-wide":
        cmd.extend(["--system", "--disallow-exit","--disallow-module-loading"])

    try:
        #This may mean it's already running, but hanging in some way
        try:
            subprocess.check_call(cmd,timeout=5)
        except:
            subprocess.call(['killall','-9','pulseaudio'],timeout=5,stderr=subprocess.DEVNULL,stdout=subprocess.DEVNULL)
            subprocess.check_call(cmd,timeout=5)
            pass
        time.sleep(0.1)
        try:
            if sharePulse== "localhost":
                subprocess.check_call("pactl load-module module-native-protocol-tcp auth-ip-acl=127.0.0.1",timeout=5,shell=True)
            elif sharePulse== "network":
                subprocess.check_call("pactl load-module module-native-protocol-tcp auth-ip-acl=127.0.0.1;192.168.0.0/24;10.0.0.0/24 auth-anonymous=1",timeout=5,shell=True)
                subprocess.check_call("pactl load-module module-zeroconf-publish",timeout=5,shell=True)
        except:
            log.exception("Error configuring pulseaudio sharing")
    
    except:
        log.exception("Error configuring pulseaudio")

def _ensureConnections(*a,**k):
    "Auto restore connections in the connection list"
    try:
        with lock:
            for i in list(allConnections.keys()):
                try:
                    allConnections[i].reconnect()
                except:
                    print(traceback.format_exc())
    except:
        log.exception("Probably just a weakref that went away.")



messagebus.subscribe("/system/jack/newport",_ensureConnections)

import weakref
allConnections=weakref.WeakValueDictionary()

activeConnections=weakref.WeakValueDictionary()

#Things as they actually are
realConnections ={}

_realConnections={}

realConnectionsLock=threading.RLock()

def findReal():
    with lock:
        p = _jackclient.get_ports(is_output=True)
        with realConnectionsLock:
            _realConnections.clear()
            for i in p:
                try:
                    for j in _jackclient.get_all_connections(i):
                        _realConnections[i.name, j.name]=True
                except:
                    log.exception("Err")
            global realConnections
            realConnections=_realConnections.copy()



errlog = []

class MonoAirwire():
    """Represents a connection that should always exist as long as there
    is a reference to this object. You can also enable and disable it with 
    the connect() and disconnect() functions.
    
    They start out in the connected state
    """
    def __init__(self, orig, to):
        self.orig=orig
        self.to = to
        self.active = True

    def disconnect(self):
        self.disconnected = True
        try:
            del allConnections[self.orig, self.to]
        except:
            pass
        try:
            with lock:
                if isConnected(self.orig,self.to):
                    disconnect(self.orig, self.to)
                    del activeConnections[self.orig,self.to]
                    try:
                        with realConnectionsLock:
                            del _realConnections[self.orig,self.to]
                            global realConnections
                            realConnections=_realConnections.copy()
                    except KeyError:
                        pass

        except:
            pass

    def __del__(self):
        #TODO: Is there any possible deadlock risk at all here?
        if self.active:
            with lock:
                if isConnected(self.orig,self.to):
                    disconnect(self.orig, self.to)

    def connect(self):
        allConnections[self.orig, self.to]= self
        activeConnections[self.orig, self.to]= self

        self.connected=True
        self.reconnect()

    def reconnect(self):
        if (self.orig, self.to) in activeConnections:
            if self.orig and self.to:
                try:
                    if not isConnected(self.orig,self.to):
                        with lock:
                            connect(self.orig,self.to)
                            with realConnectionsLock:
                                _realConnections[self.orig,self.to]=True
                                global realConnections
                                realConnections=_realConnections.copy()
                except:
                    print(traceback.format_exc())


class MultichannelAirwire(MonoAirwire):
    "Link all outputs of f to all inputs of t, in sorted order"


    def _getEndpoints(self):
        f = self.orig
        if not f:
            return None,None
        
        t = self.to
        if not t:
            return None,None
        return f,t

    def reconnect(self):
        """Connects the outputs of channel strip(Or other JACK thing)  f to the inputs of t, one to one, until
        you run out of ports. 
        
        Note that channel strips only have the main inputs but can have sends,
        so we have to distinguish them in the regex.
        """
        if not self.active:
            return
        f,t=self._getEndpoints()
        if not f:
            return

        with lock:
            if _jackclient:
                outPorts = _jackclient.get_ports(f+":*",is_output=True,is_audio=True)
                inPorts = _jackclient.get_ports(t+":*",is_input=True,is_audio=True)
                #Connect all the ports
                for i in zip(outPorts,inPorts):
                    if not isConnected(i[0].name,i[1].name):
                        connect(i[0],i[1])
                        with realConnectionsLock:
                            _realConnections[i[0].name,i[1].name]=True
                            realConnections=_realConnections.copy()

    def disconnect(self):
        if not _jackclient:
            return
        f,t=self._getEndpoints()
        if not f:
            return

        with lock:
            outPorts = _jackclient.get_ports(f+":*",is_output=True,is_audio=True)
            inPorts = _jackclient.get_ports(t+":*",is_input=True,is_audio=True)

            #Connect all the ports
            for i in zip(outPorts,inPorts):
                if isConnected(i[0],i[1]):
                    disconnect(i[0],i[1])
                    try:
                        del activeConnections[i[0].name,i[1].name]
                    except KeyError:
                        pass

    def __del__(self):
        self.disconnect()


class CombiningAirwire(MultichannelAirwire):
    def reconnect(self):
        """Connects the outputs of channel strip f to the port t. As in all outputs
        to one input. If the destination is a client, connect all channnels of src to all of dest.
        """
        if not self.active:
            return
        f,t=self._getEndpoints()
        if not f:
            return
        if lock.acquire(timeout=10):
            try:
                outPorts = _jackclient.get_ports(f+":*",is_output=True,is_audio=True)
                
                if not ":" in t:
                    t+=":*"
                inPorts = _jackclient.get_ports(t,is_input=True,is_audio=True)
            

                #Connect all the ports
                for i in outPorts:
                    for j in inPorts:
                        if not isConnected(i.name,j.name):
                            connect(i,j)
                    
            finally:
                lock.release()


    def disconnect(self):
        f,t=self._getEndpoints()
        if not f:
            return

        if lock.acquire(timeout=10):
            try:
                outPorts = _jackclient.get_ports(f+":*",is_output=True,is_audio=True)
            
                if not ":" in t:
                    t+=":*"
                inPorts = _jackclient.get_ports(t)
                if not inPorts:
                    return

                #Disconnect all the ports
                for i in outPorts:
                        for j in inPorts:
                            if isConnected(i.name,j.name):
                                try:
                                    disconnect(i,j)
                                except:
                                    print(traceback.format_exc())
                                try:
                                    del activeConnections[i.name,j.name]
                                except KeyError:
                                    pass
            finally:
                lock.release()


def Airwire(f,t,forceCombining=False):

    #Can't connect to nothing, for now lets use a hack and make these nonsense
    #names so emoty strings don't connect to stuff
    if not f or not t:
        f="jdgdsjfgkldsf"
        t="dsfjgjdsfjgkl"
    if forceCombining:
        return CombiningAirwire(f,t) 
    elif f==None or t==None:
        return MonoAirwire(None,None)
    elif ":" in f:
        if not ":" in t:
           return CombiningAirwire(f,t)
        return MonoAirwire(f,t)
    else:
        return MultichannelAirwire(f,t)

def onPortConnect(a,b,connected):
    #Whem things are manually disconnected we don't
    #Want to always reconnect every time
    global realConnections

    if connected:
        with realConnectionsLock:
            if a.is_output:
                _realConnections[a.name, b.name]=True
            else:
                _realConnections[b.name, a.name]=True
        
            realConnections=_realConnections.copy()

    if not connected:
        i = (a.name,b.name)
        with realConnectionsLock:
            if (a.name,b.name) in _realConnections:
                try:
                    del _realConnections[a.name,b.name]
                except KeyError:
                    pass

            if (b.name,a.name) in _realConnections:
                try:
                    del _realConnections[b.name,a.name]
                except KeyError:
                    pass
            
            realConnections=_realConnections.copy()

        #Try to stop whatever airwire or set therof
        #from remaking the connection
        if i in activeConnections:
            try:
                #Deactivate first, that must keep it from using the api
                #From within the callback
                activeConnections[i].active=False
                del allConnections[i]
                del activeConnections[i]
            except:
                pass

        # def f():
        #     if not connected:
        #         log.debug("JACK port "+ a.name+" disconnected from "+b.name)
        #     else:
        #         log.debug("JACK port "+ a.name+" connected to "+b.name)

        # workers.do(f)

class PortInfo():
    def __init__(self, name,isInput,sname):
        self.isOutput = not isInput
        self.isInput=isInput
        self.name=name
        self.shortname = sname
        self.clientName = name[:-len(":"+sname)]


def onPortRegistered(port,registered):
    try:
        global realConnections
        "Same function for register and unregister"
        if not port:
            return

        p = PortInfo(port.name, port.is_input,port.shortname)

        if registered:
            #log.debug("JACK port registered: "+port.name)
            messagebus.postMessage("/system/jack/newport",p )
        else:
            torm = []
            with realConnectionsLock:
                for i in _realConnections:
                    if i[0]==port.name or i[1]==port.name:
                        torm.append(i)
                for i in torm:
                    del _realConnections[i]
                realConnections=_realConnections.copy()

            messagebus.postMessage("/system/jack/delport",p)
    except:
        print(traceback.format_exc())

############################################################################
####################### This section manages the actual sound IO and creates jack ports



#This code runs once when the event loads. It also runs when you save the event during the test compile
#and may run multiple times when kaithem boots due to dependancy resolution
__doc__=''

oldi,oldo,oldmidis  = None,None,None



alsa_in_instances={}
alsa_out_instances ={}

failcards = {}



toretry_in = {}
toretry_out ={}


def compressnumbers(s):
    """Take a string that's got a lot of numbers and try to make something 
        that represents that number. Tries to make
        unique strings from things like usb-0000:00:14.0-2
    
    """
    n = ''
    currentnum = ''
    for i in s:
        if i in '0123456789':
            #Exclude leading zeros
            if currentnum or (not(i=='0')):
                currentnum+=i
        else:
            n+=currentnum
            currentnum=''

    return n+currentnum

  


def tryCloseFds(p):
    if not p:
        return
    try:
        p.stdout.close()
    except:
        pass
    try:
        p.stderr.close()
    except:
        pass 
    try:
        p.stdin.close()
    except:
        pass

def try_stop(p):
    try:
        p.terminate()
    except:
        pass
    tryCloseFds(p)

def closeAlsaProcess(x):
    #Why not a proper terminate?
    #It seemed to ignore that sometimes.
    x.kill()
    x.wait()
    tryCloseFds(x)




def cleanupstring(s):
    "Get rid of special characters and common redundant words that provide no info"
    x = s.replace(":0","0")
    x = s.replace(":0",":")
    x = s.replace(":0",":")

    x = s.replace(" ","").replace("\n","").replace("*","").replace("(","")
    x=x.replace(")","").replace("-","").replace(":",".").replace("Audio","")
    x=x.replace("Lpe","").replace("-","").replace("Generic","").replace("Device",'')
    return x


def enumeratePCI():
    "generate the PCI numbering that cardsfromregex needs"
    #We are going to list out all the PCI bus devices and give them numbers
    #So we can use fewer bits for that. Here we assign them all numbers hopefully
    #From 0 to 128. Scatter around by hash to at least try to stay consistent if
    #They change.
    pciNumbers = {}
    usedNumbers = {}
    if os.path.exists("/sys/bus/pci/devices"):
        for i in sorted(os.listdir("/sys/bus/pci/devices")):
            n=struct.unpack("<Q",hashlib.sha1(i.encode('utf8')).digest()[:8])[0]%128
            while n in usedNumbers:
                n+=1
            usedNumbers[n]=i
            pciNumbers[n]=i
    return pciNumbers




def assignName(locator,longname, usedFirstWords,pciNumbers):
    """Given the relevant context, return a two-word easy to remember name to the card with the given
        'locator' which is usually something like usb-0000:00:14.0-1
    """
    #Handle USB locators
    if 'usb' in locator:
        x = locator.split("-")
        if len(x)>2:
            controller = x[1]
            usbpath = x[2]

            if not controller in pciNumbers:
                #Not a PCI controller. We don't know how to enumerate these in a compact
                #Way, so we have to just trust that there is only one or two of them
                pval = 1
                accum = 0
                #Use base 8 to convert the usb path into a number. Assume that it's little endian.
                #We of course may have numbers that overflow and such, but we will deal with collisions
                #From that.
                digitNumber = 0
                for j in reversed(controller):
                    if j.lower() in '0123456789abcdef':
                        accum+=int(j,16)*pval
                        #Optimize for PCI bus stuff.  The last digit(Reversed to the first)
                        #Can only be 0 to 7. We use a nonlinear base system where the firt hex
                        #digit is multiplied by 8, and the second by 8*16, even though it can be up to 15.
                        #This is very confusing.
                        if digitNumber:
                            pval*=16
                        else:
                            pval*=8
                        digitNumber +=1

                #We just have to trust that we aren't likely to have two of these weird USB
                #things with locators separated by more than 128, and if we do the chance
                #of collision is low anyway.  This whole block is mostly for embedded ARM stuff
                #with one controller
                n = usbpath[0]
                if n in '0123456789':
                    accum = int(n)*128
            else:
                #We know the PCI number we generated for this. Take the first digit
                #Of the path and add it in, so that the first word varies with root port.
                #This seems kinda nice, to group things by root port.
                n = usbpath[0]
                if n in '0123456789':
                    accum = int(n)*128 +pciNumbers[controller]
                else:
                    accum=pciNumbers[controller]

            #Choose an unused first word based on what the controller is.
            first= wordlist[accum%1621]
            s= 0
            #Now we are going to fix collisons.
            while first in usedFirstWords and s<100:
                s+=1
                if usedFirstWords[first]==controller:
                    break
                first=memorableHash(controller+str(s))
            usedFirstWords[first]=controller

            pval = 1
            accum = 0
            #Use base 8 to convert the usb path into a number. Assume that it's little endian.
            #We of course may have numbers that overflow and such, but we will deal with collisions
            #From that.
            for j in usbpath:
                if j in '0123456789':
                    accum+=int(j)*pval
                    pval*=8
            # 1621 is a prime number. We use this for modulo
            # to prevent discarding higher order bits.
            second = wordlist[accum%1621]
        else:
            first = memorableHash(x[2])
            second = memorableHash(x[1])
    #handle nonusb
    else:
        controller =locator
        first=memorableHash(controller)

        while first in usedFirstWords and s<100:
            s+=1
            if usedFirstWords[first]==controller:
                break
            first=memorableHash(controller+str(s))
        second=memorableHash(longname)


    #Note that we could still have collisions. We are going to fix that
    #In the second step
    h = first+second
    return h

def generateJackName(words,longname, numberstring,  taken, taken2):
    "Generate a name suitable for direct use as a jack client name"
    n = cleanupstring(longname)
    n =n.lower()

    #You can probably see why using the first four letters of this would be
    #Unprofessional and not informative.
    if n.startswith("analog"):
        n = "anlg"

    jackname =n[:4]+'_'+words
    jackname+=numberstring
    jackname=jackname[:28]
    #If there's a collision, we're going to redo everything
    #This of course will mean we're going back to 
    while (jackname in taken) or jackname in taken2:
        h = memorableHash(jackname+cards[i[0]]+":"+i[2])[:12]
        n = cleanupstring(longname)
        jackname =n[:4]+'_'+words
        jackname+=numberstring
        jackname=jackname[:28]
    return jackname

def cardsfromregex(m, cards,usednames = [],pciNumbers={},usedFirstWords = {}):
    """Given the regex matches from arecord or aplay, match them up to the actual 
    devices and give them memorable aliases"""

    d = {}
    #Why sort? We need consistent ordering so that our conflict resolution
    #has the absolute most possible consistency
    def key(a):
        "Sort USBs last, so that PCI stuff gets the first pick"
        return(1 if ('usb' in a[2].lower()) else 0, a)

    m= sorted(m)
    for i in m:
        #Filter out HDMI
        if 'HDMI' in i[3]:
            continue
        #We generate a name that contains both the device path and subdevice
        generatedName = cards[i[0]]+"."+i[2]

        longname=i[3]
        locator = cards[i[0]]
        numberstring = compressnumbers(locator)

        h = assignName(locator,longname,usedFirstWords,pciNumbers)

        jackname = generateJackName(h, longname,numberstring,d, usednames)
        
        longname = h+"_"+memorableHash(cards[i[0]]+":"+i[2],num=4,caps=True)+"_"+cards[i[0]]
        
        try:
            d[jackname]  = ("hw:"+i[0]+","+i[2], cards[i[0]], (int(i[0]), int(i[2])), longname)
        except KeyError:
            d[jackname] = ("hw:"+i[0]+","+i[2], cards[i[0]], (int(i[0]), int(i[2])), longname)
    return d



def midiClientsToCardNumbers():
    #Groups are clientnum, name, cardnumber, only if the client actually has a card number
    pattern =r"client\s+(\d+):.+?'(.+?)'.+?card=(\d+)."
    x = subprocess.check_output(['aconnect','-i'],stderr=subprocess.DEVNULL).decode("utf8")
    x = re.findall(pattern,x)

    clients = {}
    for i in x:
        clients[int(i[0])] = (i[1], int(i[2]))
    
    x = subprocess.check_output(['aconnect','-o'],stderr=subprocess.DEVNULL).decode("utf8")
    x = re.findall(pattern,x)

    clients = {}
    for i in x:
        clients[int(i[0])] = (i[1], int(i[2]))
    return clients



midip =None

def scanMidi(m, cards,usednames = [],pciNumbers={},usedFirstWords = {}):
    """Given the regex matches from amidi -l, match them up to the actual 
    devices and give them memorable aliases.
    
    Return a dict that maps the name we have assigned the client, to a name, alsaClientNumber
    which can be used to find the ports "a2jmidid -e" may have registered, and give them aliases.

    Note that the client may have multiple ports, so we have to add disambiguation later.
    """

    d = {}

    midiInfo = midiClientsToCardNumbers()

    for i in midiInfo:
        name= midiInfo[i][0]
        card = str(midiInfo[i][1])
        
        if card in cards:
            locator = cards[card]
        else:
            continue

        numberstring = compressnumbers(locator)

        h = assignName(locator,name,usedFirstWords,pciNumbers)

        jackname = generateJackName(h, name,numberstring,d, usednames)
        
        d[jackname]  = (name,i)
    return d



def readAllSoFar(proc, retVal=b''): 
  counter = 1024
  while counter:
    x =(select.select([proc.stdout],[],[],0.1)[0])
    if x:   
        retVal+=proc.stdout.read(1)
    else:
        break
    counter -=1
  return retVal

def readAllErrSoFar(proc, retVal=b''): 
  counter = 1024
  while counter:
    x =(select.select([proc.stderr],[],[],0.1)[0])
    if x:   
        retVal+=proc.stderr.read(1)
    else:
        break
    counter -=1
  return retVal



def getPersistentCardNames():
    "Get a simple dict mapping persistent name to ALSA name, for all cards"
    i,o,x = listSoundCardsByPersistentName()

    d = {}

    for c in i:
        d[c] = i[c][0]
    for c in o:
        d[c] = o[c][0]

    return d


def listSoundCardsByPersistentName():
    """
        Only works on linux or maybe mac

       List devices in a dict indexed by human-readable and persistant easy to memorize
       Identifiers. Output is tuples:
       (cardnamewithsubdev(Typical ASLA identifier),physicalDevice(persistent),devicenumber, subdevice)
    
    """
    with open("/proc/asound/cards") as f:
        d = f.read()

    #RE pattern produces cardnumber, cardname, locator
    c = re.findall(r"[\n\r]*\s*(\d)+\s*\[(\w+)\s*\]:\s*.*?[\n\r]+\s*.*? at (.*?)[\n\r]+",d)

    #Catch the ones that don't have an "at"
    c2 = re.findall(r"[\n\r]*\s*(\d)+\s*\[(\w+)\s*\]:\s*.*?[\n\r]+\s*(.*?)[\n\r]+",d)

    cards = {}
    #find physical cards
    for i in c:
        n = i[2].strip().replace(" ","").replace(",fullspeed","").replace("[","").replace("]","")
        cards[i[0]] = n

    #find physical cards
    for i in c2:
        #Ones with at are caught in c
        if ' at ' in i[2]:
            continue
        n = i[2].strip().replace(" ","").replace(",fullspeed","").replace("[","").replace("]","")
        cards[i[0]] = n

    pci = enumeratePCI()
    usedFirstWords ={}

    x = subprocess.check_output(['aplay','-l'],stderr=subprocess.DEVNULL).decode("utf8")
    #Groups are cardnumber, cardname, subdevice, longname
    sd = re.findall(r"card (\d+): (\w*)\s\[.*?\], device (\d*): (.*?)\s+\[.*?]",x)
    outputs= cardsfromregex(sd,cards,pciNumbers=pci,usedFirstWords=usedFirstWords)
   

    x = subprocess.check_output(['arecord','-l'],stderr=subprocess.DEVNULL).decode("utf8")
    #Groups are cardnumber, cardname, subdevice, longname
    sd = re.findall(r"card (\d+): (\w*)\s\[.*?\], device (\d*): (.*?)\s+\[.*?]",x)
    inputs=cardsfromregex(sd,cards, pciNumbers=pci,usedFirstWords=usedFirstWords)

    midis=scanMidi(sd,cards, pciNumbers=pci,usedFirstWords=usedFirstWords)


    return inputs,outputs,midis


def memorableHash(x, num=1, separator="",caps=False):
    "Use the diceware list to encode a hash. Not meant to be secure."
    o = ""

    if isinstance(x, str):
        x = x.encode("utf8")
    for i in range(num):
        while 1:
            x = hashlib.sha256(x).digest()
            n = struct.unpack("<Q",x[:8])[0]%len(wordlist)
            e = wordlist[n]
            if caps:
                e=e.capitalize()
            #Don't have a word that starts with the letter the last one ends with
            #So it's easier to read
            if o:
                if e[0] == o[-1]:
                    continue
                o+=separator+e
            else:
                o=e
            break
    return o



def shortHash(x, num=8, separator=""):
    "Use the diceware list to encode a hash. Not meant to be secure."
    o = ""

    if isinstance(x, str):
        x = x.encode("utf8")
    x = hashlib.sha256(x).digest()
    b = base64.b64encode(x).decode("utf8").replace("+","").replace("/","")
    return b[:num]



def cleanup():
    import subprocess
    try:
        print("Killing jack process")
        jackp.kill()
    except:
        pass
    with lock:
        for i in alsa_in_instances:
            alsa_in_instances[i].terminate()
        for i in alsa_out_instances:
            alsa_out_instances[i].terminate()
atexit.register(cleanup)

jackShouldBeRunning = False


def stopJackServer():
    with lock:
        global jackShouldBeRunning
        jackShouldBeRunning =False
        
        _stopJackProcess()

jackp = None
lastJackStartAttempt=0




def _stopJackProcess():
    global _jackclient,jackp
    import subprocess
    from . import fluidsynth
    log.info("Stopping JACK and all related processes")
    #Get rid of old stuff
    iceflow.stopAllJackUsers()
    fluidsynth.stopAll()

    try:
        subprocess.call(['killall','alsa_in'],stderr=subprocess.DEVNULL,stdout=subprocess.DEVNULL)
    except:
        print(traceback.format_exc())
    try:
        subprocess.call(['killall','alsa_out'],stderr=subprocess.DEVNULL,stdout=subprocess.DEVNULL)
    except:
        print(traceback.format_exc())
    try:
        subprocess.call(['killall','a2jmidid'],stderr=subprocess.DEVNULL,stdout=subprocess.DEVNULL)
    except:
        print(traceback.format_exc())

    try:
        subprocess.call(['killall', 'jackd'],stderr=subprocess.DEVNULL,stdout=subprocess.DEVNULL)
    except:
        logging.exception("Failed to stop, retrying with -9")

    time.sleep(1)

    try:
        subprocess.call(['killall','-9', 'jackd'],stderr=subprocess.DEVNULL,stdout=subprocess.DEVNULL)
    except:
        print(traceback.format_exc())
    try:
        #Close any existing stuff
        if _jackclient:
            _jackclient.close()
            _jackclient=None
    except:
        print(traceback.format_exc())
        log.exception("Probably just already closed")
    tryCloseFds(jackp)
    jackp=None




def _startJackProcess(p=None, n=None,logErrs=True):
    #Start the JACK server, and related processes
    global jackp
    global periodSize
    global jackPeriods
    global jackDevice
    global currentPrimaryDevice 

    period = p or periodSize
    jackPeriods = n or jackPeriods

    log.info("Attempting to start JACK")
    if not jackp or not jackp.poll()==None:
        tryCloseFds(jackp)
        jackp=None
        if midip:
            try:
                midip.kill()
            except:
                print(traceback.format_exc())

        try:
            subprocess.call(['pulseaudio','-k'],stderr=subprocess.DEVNULL,stdout=subprocess.DEVNULL)
        except:
            print(traceback.format_exc())

        #Let's be real sure it's gone
        try:
            subprocess.call(['killall','-9', 'pulseaudio'],stderr=subprocess.DEVNULL,stdout=subprocess.DEVNULL)
        except:
            print(traceback.format_exc())

        #TODO: Explicitly close all the FDs we open!
        f = open(os.devnull,"w")
        my_env = os.environ.copy()
 
        if not 'DBUS_SESSION_BUS_ADDRESS' in my_env or not my_env['DBUS_SESSION_BUS_ADDRESS']:
            my_env['DBUS_SESSION_BUS_ADDRESS'] = "unix:path=/run/dbus/system_bus_socket"
            os.environ['DBUS_SESSION_BUS_ADDRESS'] ="1"

        os.environ['JACK_NO_AUDIO_RESERVATION'] ="1"
        my_env['JACK_NO_AUDIO_RESERVATION'] ="1"

        settingsReloader()

        lastJackStartAttempt=time.monotonic()

        useDevice = "hw:0,0"
        
        #If the user supplied a perisistent name, translate it.
        if jackDevice:
            incards,outcards, x= listSoundCardsByPersistentName()
            if jackDevice in incards:
                useDevice = incards[jackDevice][0]

            if jackDevice in outcards:
                useDevice = outcards[jackDevice][0]
        
        currentPrimaryDevice = useDevice

        if realtimePriority:
            cmdline = ['jackd','-S', '--realtime', '-P' ,str(realtimePriority) ,'-d', 'alsa' ,'-d' ,useDevice ,'-s','-p' ,str(period), '-n' ,str(jackPeriods) ,'-r','48000']
        else:
            cmdline = ['jackd', '-S', '--realtime' ,'-d', 'alsa' ,'-d' ,useDevice ,'-p' ,str(period),'-s', '-n' ,str(jackPeriods) ,'-r','48000']
        logging.info("Attempting to start JACKD server with: \n"+' '.join(cmdline))

        jackp = subprocess.Popen(cmdline,stdin=subprocess.DEVNULL,stdout=subprocess.PIPE, stderr=subprocess.PIPE,env=my_env)
        
        time.sleep(0.5)

        if jackp.poll() != None:
            x = readAllErrSoFar(jackp)
            if x and logErrs:
                log.error("jackd:\n"+x.decode('utf8','ignore'))
            x=readAllSoFar(jackp)
            if x and logErrs:
                log.info("jackd:\n"+x.decode('utf8','ignore'))
            tryCloseFds(jackp)
            jackp=None


        if jackp and jackp.poll() != None:
            tryCloseFds(jackp)
            jackp=None
            raise RuntimeError("Could not start JACK process")

        
        startedClient = False

        #Try to start the JACK client
        for i in range(0,10):
            try:
                if _checkJackClient(err=False):
                    startedClient=True
                    break
            except:
                time.sleep(1)
        global prevJackStatus

        #Try again, but this time don't just hide the error.
        if not startedClient:
            if not _checkJackClient():
                if prevJackStatus:
                    prevJackStatus=False
                    onJackFailure()
        else:
            if not prevJackStatus:
                prevJackStatus=True
                onJackStart()
            logging.debug("Connected to JACKD server")



        
        if realtimePriority and jackp:
            try:
                subprocess.check_call(['chrt', '-f','-p', str(realtimePriority), str(jackp.pid)])
            except:
                log.exception("Error getting RT")
      
        try:
            setupPulse()
            log.debug("Set up pulse")
        except:
            log.exception("Error starting pulse, ignoring")    
        def f():
            global midip
            #Poll till it's actually started, then post the message
            for i in range(120):
                if getPorts():
                    break
                time.sleep(0.5)
            messagebus.postMessage("/system/jack/started",{})
           
            if util.which("a2jmidid"):
                midip = subprocess.Popen("a2jmidid -e",stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, shell=True,stdin=subprocess.DEVNULL) 
            else:
                logging.error("a2jmidid not installed, MIDI may not work as expected.")

            from . import fluidsynth
            fluidsynth.remakeAll()

        workers.do(f)




def getIOOptionsForAdditionalSoundcard():
    settingsReloader()
    psize = usbPeriodSize
    iooptions=[]
    
    if not psize<0:
        iooptions+=["-p",str(psize)]

    if not usbLatency<0:
        iooptions+=["-t",str(usbLatency)]

    if not usbPeriods<0:
        iooptions+=["-n",str(usbPeriods)]

    iooptions+=["-q",str(usbQuality), "-r", "48000"]

    return iooptions

lastFullScan=0

def handleManagedSoundcards():
    "Make sure that all of our alsa_in and out instances are working as they should be."
    global oldi
    global oldo
    global oldmidis
    global lastFullScan
    global useAdditionalSoundcards

    #There seems to be a bug in reading errors from the process
    #Right now it's a TODO, but most of the time
    #we catch things in the add/remove detection anyway
    if lock.acquire(timeout=10):
        try:     
            try:
                tr =[]
                for i in alsa_out_instances:
                    
                    x=readAllSoFar(alsa_out_instances[i])
                    e=readAllErrSoFar(alsa_out_instances[i])
                    problem = b"err =" in x+e or alsa_out_instances[i].poll()
                    problem = problem or b"busy" in (x+e) 

                    if problem:
                        log.error("Error in output "+ i +(x+e).decode("utf8") +" status code "+str(alsa_out_instances[i].poll()))
                        closeAlsaProcess(alsa_out_instances[i])
                        tr.append(i)
                        #We have to delete the busy stuff but we can
                        #retry later
                        if b"busy" in (x+e):
                            toretry_out[i]=time.monotonic()+5
                        elif b"No such" in (x+e):
                            toretry_out[i]=time.monotonic()+10
                            log.error("Nonexistant card "+i)
                        else:
                            toretry_out[i]=time.monotonic()

                        log.info("Removed "+i+"o")

                    elif not alsa_out_instances[i].poll()==None:
                        tr.append(i)
                        log.info("Removed "+i+"o")

                for i in tr:
                    try_stop(alsa_out_instances[i])
                    del alsa_out_instances[i]

                tr =[]
                for i in alsa_in_instances:
                    
                    x= readAllSoFar(alsa_in_instances[i])
                    e=readAllErrSoFar(alsa_in_instances[i])
                    problem = b"err =" in x+e or alsa_in_instances[i].poll()
                    problem = problem or b"busy" in (x+e) 
                    if problem :
                        log.error("Error in input "+ i +(x+e).decode("utf8") +" status code "+str(alsa_in_instances[i].poll()))
                        closeAlsaProcess(alsa_in_instances[i])   
                        tr.append(i)
                        if b"busy" in (x+e):
                            toretry_in[i]=time.monotonic()+5
                        
                        if b"No such" in (x+e):
                            toretry_in[i]=time.monotonic()+10
                            log.error("Nonexistant card "+i)
                        else:
                            toretry_out[i]=time.monotonic()

                        log.info("Removed "+i+"i")

                    elif not alsa_in_instances[i].poll()==None:
                        tr.append(i)
                        log.info("Removed "+i+"i")

                for i in tr:
                    try_stop(alsa_in_instances[i])
                    del alsa_in_instances[i]
            except:
                print(traceback.format_exc())
                



            ##HANDLE CREATING AND GC-ING things
            inp,op,midis = listSoundCardsByPersistentName()

            #Ignore all except MIDI
            if not useAdditionalSoundcards.lower() in ("true","yes",'on'):
                inp = {}
                op = {}

            #This is how we avoid constantky retrying to connect the same few
            #clients that fail, which might make a bad periodic click that nobody
            #wants to hear.
            startPulse = False
            if (inp,op,midis)==(oldi,oldo,oldmidis):

                #However some things we need to retry.
                #Device or resource busy being the main one
                for i in inp:
                    if i in toretry_in:
                        if time.monotonic() < toretry_in[i]:
                            continue
                        del toretry_in[i]
                        if not i in alsa_in_instances:
                            #Pulse likes to take over cards so we have to stop it, take the card, then start it. It sucks
                            startPulse = True
                            try:
                                subprocess.call(['pulseaudio','-k'],stderr=subprocess.DEVNULL,stdout=subprocess.DEVNULL)
                            except:
                                print(traceback.format_exc())
                            time.sleep(2)

                            log.debug("Starting alsa_in process, for "+inp[i][0])
                            x = subprocess.Popen(["alsa_in"]+getIOOptionsForAdditionalSoundcard()+["-d", inp[i][0], "-j",i+"i"],stdout=subprocess.PIPE,stderr=subprocess.PIPE)
                            try:
                                subprocess.check_call(['chrt', '-f','-p', '70', str(x.pid)])
                            except:
                                log.exception("Error getting RT")
        
                            alsa_in_instances[i]=x
                            log.info("Added "+i+"i at"+inp[i][1])

                for i in op:
                    if i in toretry_out:
                        if time.monotonic() < toretry_out[i]:
                            continue
                        del toretry_out[i]
                        if not i in alsa_out_instances:
                            startPulse=True
                            try:
                                subprocess.call(['pulseaudio','-k'],stderr=subprocess.DEVNULL,stdout=subprocess.DEVNULL)
                            except:
                                print(traceback.format_exc())

                            log.debug("Starting alsa_out process for "+op[i][0])
                            x = subprocess.Popen(["alsa_out"]+getIOOptionsForAdditionalSoundcard()+["-d", op[i][0], "-j",i+"o"],stdout=subprocess.PIPE,stderr=subprocess.PIPE)
                            alsa_out_instances[i]=x

                            try:
                                subprocess.check_call(['chrt', '-f','-p', '70', str(x.pid)])
                            except:
                                log.exception("Error getting RT")
                            log.info("Added "+i+"o")
                
                #If we stopped it, start it again
                if startPulse:
                    try:
                        setupPulse()
                    except:
                        log.exception("Error restarting pulse, ignoring")
                if lastFullScan> time.monotonic()-10:
                    return
                lastFullScan = time.monotonic()
            oldi,oldo,oldmidis =inp,op,midis

            #Look for ports with the a2jmidid naming pattern and give them persistant name aliases.
            x = _jackclient.get_ports(is_midi=True)
            for i in midis:
                for j in x:
                    number = "["+str(midis[i][1])+"]"
                    if number in j.name:
                        if i[0] in j.name:
                            try:
                                if not i in j.aliases:
                                    j.set_alias(i)
                            except:
                                log.exception("Error setting MIDI alias")


            for i in inp:
                #HDMI doesn't do inputs as far as I know
                if not i.startswith("HDMI"):
                    if not i in alsa_in_instances:
                        if inp[i][0]== currentPrimaryDevice:
                            #We don't do an alsa in for this card because it
                            #Is already the JACK backend
                            if usingDefaultCard:
                                continue

                        log.debug("Starting alsa_in process for "+inp[i][0])
                        x = subprocess.Popen(["alsa_in"]+getIOOptionsForAdditionalSoundcard()+["-d", inp[i][0], "-j",i+"i"],stdout=subprocess.PIPE,stderr=subprocess.PIPE)
                    
                        try:
                            subprocess.check_call(['chrt', '-f','-p', '70', str(x.pid)])
                        except:
                            log.exception("Error getting RT")
                        alsa_in_instances[i]=x
                        log.info("Added "+i+"i at "+inp[i][1])

            for i in op:
                if not i.startswith("HDMI"):
                    if not i in alsa_out_instances:
                        #We do not want to mess with the 
                        if op[i][0]== currentPrimaryDevice:
                            if usingDefaultCard:
                                continue
                        
                        log.debug("Starting alsa_out process, for "+op[i][0])
                        x = subprocess.Popen(["alsa_out"]+getIOOptionsForAdditionalSoundcard()+["-d", op[i][0], "-j",i+'o'],stdout=subprocess.PIPE,stderr=subprocess.PIPE)
                    
                        try:
                            subprocess.check_call(['chrt', '-f','-p', '70', str(x.pid)])
                        except:
                            log.exception("Error getting RT")
                        alsa_out_instances[i]=x
                        log.info("Added "+i+"o at "+op[i][1])

            #In case alsa_in doesn't properly tell us about a removed soundcard
            #Check for things that no longer exist.
            try:
                tr =[]
                for i in alsa_out_instances:
                    if not i in op:
                        tr.append(i)
                for i in tr:
                    log.warning("Removed "+i+"o because the card was removed")
                    try:
                        closeAlsaProcess(alsa_out_instances[i])
                    except:
                        log.exception("Error closing process")
                    del alsa_out_instances[i]

                tr =[]
                for i in alsa_in_instances:
                    if not i in inp:
                        tr.append(i)
                for i in tr:
                    log.warning("Removed "+i+"i because the card was removed")
                    try:
                        closeAlsaProcess(alsa_in_instances[i])
                    except:
                        log.exception("Error closing process")
                    del alsa_in_instances[i]
            except:
                log.exception("Exception in loop")
            try:
                tr =[]
                for i in alsa_out_instances:
                    
                    x=readAllSoFar(alsa_out_instances[i])
                    e=readAllErrSoFar(alsa_out_instances[i])
                    problem = b"err =" in x+e or alsa_out_instances[i].poll()
                    problem = problem or b"busy" in (x+e) 

                    if problem:
                        log.error("Error in output "+ i +(x+e).decode("utf8") +" status code "+str(alsa_out_instances[i].poll()))
                        closeAlsaProcess(alsa_out_instances[i])
                        tr.append(i)
                        #We have to delete the busy stuff but we can
                        #retry later
                        if b"busy" in (x+e):
                            toretry_out[i]=time.monotonic()+5
                        elif b"No such" in (x+e):
                            toretry_out[i]=time.monotonic()+10
                            log.error("Nonexistant card "+i)
                        else:
                            toretry_out[i]=time.monotonic()

                        log.info("Removed "+i+"o")

                    elif not alsa_out_instances[i].poll()==None:
                        tr.append(i)
                        log.info("Removed "+i+"o")

                for i in tr:
                    try_stop(alsa_out_instances[i])
                    del alsa_out_instances[i]

                tr =[]
                for i in alsa_in_instances:
                    
                    x= readAllSoFar(alsa_in_instances[i])
                    e=readAllErrSoFar(alsa_in_instances[i])
                    problem = b"err =" in x+e or alsa_in_instances[i].poll()
                    problem = problem or b"busy" in (x+e) 
                    if problem :
                        log.error("Error in input "+ i +(x+e).decode("utf8") +" status code "+str(alsa_in_instances[i].poll()))
                        closeAlsaProcess(alsa_in_instances[i])   
                        tr.append(i)
                        if b"busy" in (x+e):
                            toretry_in[i]=time.monotonic()+5
                        
                        if b"No such" in (x+e):
                            toretry_in[i]=time.monotonic()+10
                            log.error("Nonexistant card "+i)
                        else:
                            toretry_out[i]=time.monotonic()

                        log.info("Removed "+i+"i")

                    elif not alsa_in_instances[i].poll()==None:
                        tr.append(i)
                        log.info("Removed "+i+"i")

                for i in tr:
                    try_stop(alsa_in_instances[i])
                    del alsa_in_instances[i]
            except:
                print(traceback.format_exc())
                



            ##HANDLE CREATING AND GC-ING things
            inp,op,midis = listSoundCardsByPersistentName()

            #Ignore all except MIDI
            if not useAdditionalSoundcards.lower() in ("true","yes",'on'):
                inp = {}
                op = {}

            #This is how we avoid constantky retrying to connect the same few
            #clients that fail, which might make a bad periodic click that nobody
            #wants to hear.
            startPulse = False
            if (inp,op,midis)==(oldi,oldo,oldmidis):

                #However some things we need to retry.
                #Device or resource busy being the main one
                for i in inp:
                    if i in toretry_in:
                        if time.monotonic() < toretry_in[i]:
                            continue
                        del toretry_in[i]
                        if not i in alsa_in_instances:
                            #Pulse likes to take over cards so we have to stop it, take the card, then start it. It sucks
                            startPulse = True
                            try:
                                subprocess.call(['pulseaudio','-k'],stderr=subprocess.DEVNULL,stdout=subprocess.DEVNULL)
                            except:
                                print(traceback.format_exc())
                            time.sleep(2)

                            log.debug("Starting alsa_in process, for "+inp[i][0])
                            x = subprocess.Popen(["alsa_in"]+getIOOptionsForAdditionalSoundcard()+["-d", inp[i][0], "-j",i+"i"],stdout=subprocess.PIPE,stderr=subprocess.PIPE)
                            try:
                                subprocess.check_call(['chrt', '-f','-p', '70', str(x.pid)])
                            except:
                                log.exception("Error getting RT")
        
                            alsa_in_instances[i]=x
                            log.info("Added "+i+"i at"+inp[i][1])

                for i in op:
                    if i in toretry_out:
                        if time.monotonic() < toretry_out[i]:
                            continue
                        del toretry_out[i]
                        if not i in alsa_out_instances:
                            startPulse=True
                            try:
                                subprocess.call(['pulseaudio','-k'],stderr=subprocess.DEVNULL,stdout=subprocess.DEVNULL)
                            except:
                                print(traceback.format_exc())

                            log.debug("Starting alsa_out process for "+op[i][0])
                            x = subprocess.Popen(["alsa_out"]+getIOOptionsForAdditionalSoundcard()+["-d", op[i][0], "-j",i+"o"],stdout=subprocess.PIPE,stderr=subprocess.PIPE)
                            alsa_out_instances[i]=x

                            try:
                                subprocess.check_call(['chrt', '-f','-p', '70', str(x.pid)])
                            except:
                                log.exception("Error getting RT")
                            log.info("Added "+i+"o")
                
                #If we stopped it, start it again
                if startPulse:
                    try:
                        setupPulse()
                    except:
                        log.exception("Error restarting pulse, ignoring")
                if lastFullScan> time.monotonic()-10:
                    return
                lastFullScan = time.monotonic()
            oldi,oldo,oldmidis =inp,op,midis

            #Look for ports with the a2jmidid naming pattern and give them persistant name aliases.
            x = _jackclient.get_ports(is_midi=True)
            for i in midis:
                for j in x:
                    number = "["+str(midis[i][1])+"]"
                    if number in j.name:
                        if i[0] in j.name:
                            try:
                                if not i in j.aliases:
                                    j.set_alias(i)
                            except:
                                log.exception("Error setting MIDI alias")


            for i in inp:
                #HDMI doesn't do inputs as far as I know
                if not i.startswith("HDMI"):
                    if not i in alsa_in_instances:
                        if inp[i][0]== currentPrimaryDevice:
                            #We don't do an alsa in for this card because it
                            #Is already the JACK backend
                            if usingDefaultCard:
                                continue

                        log.debug("Starting alsa_in process for "+inp[i][0])
                        x = subprocess.Popen(["alsa_in"]+getIOOptionsForAdditionalSoundcard()+["-d", inp[i][0], "-j",i+"i"],stdout=subprocess.PIPE,stderr=subprocess.PIPE)
                    
                        try:
                            subprocess.check_call(['chrt', '-f','-p', '70', str(x.pid)])
                        except:
                            log.exception("Error getting RT")
                        alsa_in_instances[i]=x
                        log.info("Added "+i+"i at "+inp[i][1])

            for i in op:
                if not i.startswith("HDMI"):
                    if not i in alsa_out_instances:
                        #We do not want to mess with the 
                        if op[i][0]== currentPrimaryDevice:
                            if usingDefaultCard:
                                continue
                        
                        log.debug("Starting alsa_out process, for "+op[i][0])
                        x = subprocess.Popen(["alsa_out"]+getIOOptionsForAdditionalSoundcard()+["-d", op[i][0], "-j",i+'o'],stdout=subprocess.PIPE,stderr=subprocess.PIPE)
                    
                        try:
                            subprocess.check_call(['chrt', '-f','-p', '70', str(x.pid)])
                        except:
                            log.exception("Error getting RT")
                        alsa_out_instances[i]=x
                        log.info("Added "+i+"o at "+op[i][1])

            #In case alsa_in doesn't properly tell us about a removed soundcard
            #Check for things that no longer exist.
            try:
                tr =[]
                for i in alsa_out_instances:
                    if not i in op:
                        tr.append(i)
                for i in tr:
                    log.warning("Removed "+i+"o because the card was removed")
                    try:
                        closeAlsaProcess(alsa_out_instances[i])
                    except:
                        log.exception("Error closing process")
                    del alsa_out_instances[i]

                tr =[]
                for i in alsa_in_instances:
                    if not i in inp:
                        tr.append(i)
                for i in tr:
                    log.warning("Removed "+i+"i because the card was removed")
                    try:
                        closeAlsaProcess(alsa_in_instances[i])
                    except:
                        log.exception("Error closing process")
                    del alsa_in_instances[i]
            except:
                log.exception("Exception in loop")
        finally:
            lock.release()


def work():
    global _reconnecterThreadObjectStopper

    #Wait 10s before actually doing anything to avoid nuisiance chattering errors.
    #This thread mostly only fixes crashed stuff.
    for i in range(100):
        if not _reconnecterThreadObjectStopper[0]:
            return
        time.sleep(0.1)

    while(_reconnecterThreadObjectStopper[0]):
        try:
            #The _checkJack stuf won't block, because we already have the lock
            if lock.acquire(timeout=0.5):
                try:
                    if manageJackProcess:
                        _checkJack()
                    _checkJackClient()
                    if manageSoundcards:
                        handleManagedSoundcards()
                finally:
                    lock.release()
                _ensureConnections()
            else:
                if(_reconnecterThreadObjectStopper[0]):
                    #Might be worth logging
                    raise RuntimeError("Could not get lock,retrying in 5s")
                else:
                    #Already stopping anyway, ignore
                    pass
            time.sleep(5)
        except:
            logging.exception("Error in jack manager")
_reconnecterThreadObject =None
_reconnecterThreadObjectStopper = [0]

_didPatch = False



def _doPatch():
    import jack, re
    global _didPatch
    with lock:
        if not _didPatch:
            def _get_ports_fix(self, name_pattern='', is_audio=False, is_midi=False,
                        is_input=False, is_output=False, is_physical=False,
                        can_monitor=False, is_terminal=False):
                if name_pattern:
                    re.compile(name_pattern)
                        
                return jack.Client._get_ports(self, name_pattern, is_audio, is_midi, 
                                            is_input, is_output, is_physical, 
                                            can_monitor, is_terminal)

            jack.Client._get_ports = jack.Client.get_ports
            jack.Client.get_ports = _get_ports_fix
            _didPatch = True
            atexit.register(cleanup)





def startJackProcess(p=None,n=None):
    with lock:
        global jackShouldBeRunning
        jackShouldBeRunning = True

        for i in range(10):
            try:
                _stopJackProcess()
                #Only log on the last attempt
                _startJackProcess(p,n,logErrs=i==9)
                break
            except:
                log.exception("Failed to start JACK?")
                time.sleep(0.5)
                if i==9:
                    raise

def startManaging(p=None,n=None):
    "Start mananaging JACK in whatever way was configured."
    import jack, re
    global _jackclient
    global _reconnecterThreadObject
    _doPatch()
    
    with lock:
        #Stop the old thread if needed
        _reconnecterThreadObjectStopper[0] = 0
        try:
            if _reconnecterThreadObject:
                _reconnecterThreadObject.join()
        except:
            pass

        if manageJackProcess:
            try:
                startJackProcess()
            except:
                log.exception("Could not start JACK, retrying later")

        _reconnecterThreadObjectStopper[0]=1
        _reconnecterThreadObject = threading.Thread(target=work)
        _reconnecterThreadObject.name="JackReconnector"
        _reconnecterThreadObject.daemon=True
        _reconnecterThreadObject.start()

      

def stopManaging():
    with lock:
        #Stop the old thread if needed
        _reconnecterThreadObjectStopper[0] = 0
        try:
            if _reconnecterThreadObject:
                _reconnecterThreadObject.join()
        except:
            pass
        _reconnecterThreadObject=None
        
        if manageJackProcess:
            stopJackServer()


jack_output = b''
def _checkJack():
    global jackp,jack_output
    if lock.acquire(timeout=10):
        try:
            if manageJackProcess:
                if jackp:
                    try:
                        jack_output+= readAllSoFar(jackp)
                        jack_output+= readAllErrSoFar(jackp)
                    except:
                        print(traceback.format_exc())
                        print(jack_output)

                        try:
                            print("Killing jack process, read fail")
                            jackp.kill()
                        except:
                            pass
                        jackp = None

                    if b"\n" in jack_output or len(jack_output)>1024:
                        print("jackd:")
                        print(jack_output)
                        jack_output=b''


                if manageJackProcess  and (not jackp) or (jackp and jackp.poll() !=None):          
                    global prevJackStatus
                    if not jackp:
                        if prevJackStatus:
                            prevJackStatus=False
                            onJackFailure()
                    if time.monotonic()-lastJackStartAttempt > 10:
                        log.warning("JACK appears to have stopped. Attempting restart.")
                        _stopJackProcess()
                        if jackShouldBeRunning:
                            _startJackProcess()
        finally:
            lock.release()

def _checkJackClient(err=True):
    global _jackclient, realConnections
    import jack
    if lock.acquire(timeout=10):
        try:
             
            t=_jackclient.status.server_started
            if not t:
                raise RuntimeError("JACK Server not started or client failed")
            return True
        except:
            try:
                _jackclient.close()
                _jackclient=None
            except:
                pass
            _realConnections = {}

            try:
                _jackclient=jack.Client("Overseer",no_start_server=True)
            except:
                if err:
                    log.exception("Error creating JACK client")
                return

            _jackclient.set_port_registration_callback(onPortRegistered,only_available=True)
            _jackclient.set_port_connect_callback(onPortConnect)
            _jackclient.activate()
            _jackclient.get_ports()
            time.sleep(0.5)
            findReal()
            return True
        finally:
            lock.release()

    if not _jackclient:
        return False

def getPorts(*a,**k):
    if lock.acquire(timeout=10):
        try:
            if not _jackclient:
                return []
            ports =[]
            return _jackclient.get_ports(*a,**k)
        finally:
            lock.release()

def getPortNamesWithAliases(*a,**k):
    if lock.acquire(timeout=10):
        try:
            if not _jackclient:
                return []
            ports =[]
            x= _jackclient.get_ports(*a,**k)
            for i in x:
                for j in i.aliases:
                    if not j in ports:
                        ports.append(j)
                if not i.name in ports:
                    ports.append(i.name)
            return ports
        finally:
            lock.release()


def getConnections(name,*a,**k):
    if lock.acquire(timeout=10):
        try:
            if not _jackclient:
                return []
            try:
                return _jackclient.get_all_connections(name)
            except:
                log.exception("Error getting connections")
                return []
        finally:
            lock.release()

def disconnect(f,t):
    if iceflow.lock.acquire(timeout=30):
        try:
            if lock.acquire(timeout=30):
                try:
                    if not _jackclient:
                        return

                    if not isConnected(f,t):
                        return

                    try:
                        if  isinstance(f,str):
                            f = _jackclient.get_port_by_name(f)
                        if  isinstance(t,str):
                            t = _jackclient.get_port_by_name(t)
                
                        _jackclient.disconnect(f,t)
                    except:
                        pass
                finally:
                    lock.release()
        finally:
            iceflow.lock.release()

def connect(f,t):
    if iceflow.lock.acquire(timeout=10):
        try:
            if lock.acquire(timeout=10):
                try:
                    if not _jackclient:
                        return 
                    try:
                        if  isinstance(f,str):
                            f = _jackclient.get_port_by_name(f)
                        if  isinstance(t,str):
                            t = _jackclient.get_port_by_name(t)
                    except:
                        return

                    f_input =  f.is_input

                    if f.is_input:
                        if not t.is_output:
                            #Do a retry, there seems to be a bug somewhere
                            try:
                                f = _jackclient.get_port_by_name(f.name)
                                t = _jackclient.get_port_by_name(t.name)
                            except:
                                return
                            if f.is_input:
                                if not t.is_output:
                                    raise ValueError("Cannot connect two inputs",str((f,t)))
                    else:
                        if t.is_output:
                            raise ValueError("Cannot connect two outputs",str((f,t)))
                    f=f.name
                    t=t.name
                    try:
                        if f_input:
                            _jackclient.connect(t,f)
                        else:
                            _jackclient.connect(f,t)
                    except:
                        pass
                finally:
                    lock.release()
        finally:
            iceflow.lock.release()
#startManagingJack()





import scullery.workers
scullery.workers.start()

import scullery.iceflow
import scullery.jack

import os,time

import unittest,random,gc

class Player(scullery.iceflow.GstreamerPipeline):
    def __init__(self):
        scullery.iceflow.GstreamerPipeline.__init__(self,realtime=False)
        self.sink = self.addElement('audiotestsrc') 
        self.sink = self.addElement('jackaudiosink',client_name="JackTest") 



scullery.jack.startManaging()

#We call this manually, because we didn't set manageJackProcess
#So it won't actually auto manage the server itself
scullery.jack._startJackProcess()

class TestJackAudio(unittest.TestCase):
    def test_airwire(self):
        p=Player()
        p.start()
        
        print("You should hear noise")
        aw =scullery.jack.Airwire("JackTest","system")
        aw.connect()
        time.sleep(1)
        print("No more noise")
        
        del aw
        gc.collect()
        
        self.assertEqual(len(scullery.jack.allConnections), 0)

    def test_fluidsynth_jack(self):
        import scullery.fluidsynth
        fs = scullery.fluidsynth.FluidSynth(jackClientName="fstest",connectOutput="system")

        print("You should hear a MIDI note. This test require's fluidsynth, pyFluidsynth, and it's default sf2 file")
        fs.noteOn(0,70,50)
        time.sleep(0.5)
        fs.noteOff(0,70)
        fs.setInstrument(0,"Pizzicato")
        print("You should hear a Pizzicato string note")
        fs.noteOn(0,70,70)
        time.sleep(1)
        fs.noteOff(0,70)



    def test_z_stop(self):
        "Must obviously run last"
        print("Stopping JACK")
        scullery.jack.stopManaging()
        scullery.jack.stopJackServer()
        print("Stopped JACK")

            





